/***
 * Excerpted from "The ThoughtWorks Anthology 2",
 * published by The Pragmatic Bookshelf.
 * Copyrights apply to this code. It may not be used to create training material, 
 * courses, books, articles, and the like. Contact us if you are in doubt.
 * We make no guarantees that this code is fit for any purpose. 
 * Visit http://www.pragmaticprogrammer.com/titles/twa2 for more book information.
***/
package voterlist.rolebased;

import java.util.Collection;
import java.util.HashSet;

import static java.util.Arrays.asList;

interface VotingJurisdiction {
    boolean covers(VotingJurisdiction votingJurisdiction);
}
interface Voter {
    boolean canVoteIn(VotingJurisdiction votingJurisdiction);
}
class CouncilOfNations implements VotingJurisdiction {
    private Collection<Country> memberNations;
    public CouncilOfNations(Collection<Country> memberNations) {
        this.memberNations = memberNations;
    }
    public boolean covers(VotingJurisdiction votingJurisdiction) {
        return this.equals(votingJurisdiction) ||
            memberNations.contains(votingJurisdiction);
    }
}
class Country implements VotingJurisdiction, Voter {
    private String name;
    public Country(String name) {
        this.name = name;
    }
    public boolean covers(VotingJurisdiction votingJurisdiction) {
        return this.equals(votingJurisdiction);
    }
    public boolean canVoteIn(VotingJurisdiction votingJurisdiction) {
        return votingJurisdiction.covers(this);
    }
    public String toString() {
        return name;
    }
}
class Person implements Voter {
    private String name;
    private int age;
    private Country country;
    public Person(String name, int age, Country country) {
        // ...
        this.name = name;
        this.age = age;
        this.country = country;
    }
    public boolean canVoteIn(VotingJurisdiction votingJurisdiction) {
        return age >= 18 && votingJurisdiction.covers(country);
    }
    public String toString() {
        return name;
    }
}
class VoterList {
    private Collection<Voter> candidateVoters;
    public VoterList(Collection<Voter> candidateVoters) {
        this.candidateVoters = candidateVoters;
    }
    public Collection<Voter> votersFor(VotingJurisdiction votingJurisdiction) {
        Collection<Voter> eligibleVoters = new HashSet<Voter>();
        for (Voter voter : candidateVoters) {
            if (voter.canVoteIn(votingJurisdiction)) {
                eligibleVoters.add(voter);
            }
        }
        return eligibleVoters;
    }
}

public class VoterListRoleBased {
    public static void main(String[] args) {
        Country INDIA = new Country("India");
        Country USA = new Country("USA");
        Country UK = new Country("UK");
        Collection<Voter> persons = new HashSet<Voter>(asList(
            new Person("Donald", 28, INDIA),
            new Person("Daisy", 25, USA),
            new Person("Minnie", 17, UK)
        ));
        VoterList personVoterList = new VoterList(persons);
        System.out.println(personVoterList.votersFor(INDIA)); // [ Donald ]
        System.out.println(personVoterList.votersFor(USA));   // [ Daisy ]

        Collection<Voter> countries = new HashSet<Voter>(asList(
            INDIA, USA, UK
        ));
        VoterList countryVoterList = new VoterList(countries);
        CouncilOfNations councilOfNations = new CouncilOfNations(asList(
            USA, INDIA
        ));
        System.out.println(countryVoterList.votersFor(councilOfNations));
        // [ USA, India ]
        System.out.println(personVoterList.votersFor(councilOfNations));
        // [ Donald, Daisy ]
    }
}
